//----------------------------------------------------------------------
//  Weaver 変調器を利用する周波数シフタ，100 Hz 高い方へシフト
//
//  2023/08/21, Copyright (c) 2023 MIKAMI, Naoki
//----------------------------------------------------------------------

#include "MultirateLiPh.hpp"
#include "IirCascade.hpp"
#include "coefficientsWeaver.hpp"
#include "QuadOscIir.hpp"
#pragma diag_suppress 870   // マルチバイト文字使用の警告抑制のため
using namespace Mikami;

const float FS_ = 44.1;     // 入力の標本化周波数： 44.1 kHz
MultirateLiPh myAdDa_(FS_); // 出力標本化周波数を４倍にするオブジェクト

// 入力帯域制限用帯域通過フィルタ，通過域：0.1 ～ 19.5 kHz
IirCascade bpf_(ORDER_BPF_, C_BP_, G0_BPF_);

// Weaver 変調器で使う低域通過フィルタ，遮断周波数：9.7 kHz
IirCascade lpfC_(ORDER_LPF_, C_LP_, G0_LPF_);
IirCascade lpfS_(ORDER_LPF_, C_LP_, G0_LPF_);

const float FL_ =  0.1f;            // 入力下限の周波数： 0.1 kHz
const float FH_ = 19.5f;            // 入力上限の周波数：19.5 kHz
const float FC_ = (FL_ + FH_)/2.0f; // 帯域中央の周波数

QuadOscIir osc1_(FC_, FS_);         // 直交信号発生器，帯域中央の周波数
QuadOscIir osc2_(FC_+0.1f, FS_);    // 直交信号発生器，シフトする周波数: 0.1 kHz

bool on_ = true;

// Weaver 変調器による周波数シフトの処理
void AdcIsr()
{
    float xn = myAdDa_.Input(); // 入力

    xn = bpf_.Execute(xn);      // BPF

    float sinX1, cosX1;
    osc1_.Generate(sinX1, cosX1);   // 帯域中央の周波数の cos, sin 発生

    float mpyC = lpfC_.Execute(xn*cosX1);   // LPF
    float mpyS = lpfS_.Execute(xn*sinX1);   // LPF

    float sinX2, cosX2;
    osc2_.Generate(sinX2, cosX2);   // 0.1 kHz 高い周波数の cos, sin 発生

    float yn = 2.0f*(mpyC*cosX2 + mpyS*sinX2);

    if (on_) myAdDa_.Output(yn);    // 出力（周波数シフタの出力）
    else     myAdDa_.Output(xn);    // 出力（入力そのまま）
}

int main()
{
    setbuf(stdout, NULL);   // printf() でバッファ無しの設定
    printf("\r\nWeaver 変調器を利用する周波数シフタを実行します\r\n");
    printf("周波数シフタの ON/OFF はキーボードからコントロールできます\r\n");
    printf("'y'：ON, 'n'：Off, ただし 'y' 以外は OFF になります\r\n");
    printf("起動時は ON の状態です\r\n? ");

    myAdDa_.Start(&AdcIsr);     // 標本化を開始する
    while (true)
    {
        char yesNo = getchar();
        printf("%c\r\n? ", yesNo);
        on_ = (yesNo == 'y') ? true : false;
    }
}
